{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            copyright (c) 2020                                      }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}

unit WEBLib.Bluetooth;

interface

uses
  Classes, Web, JS, jsdelphisystem, WEBLib.Controls;

type
  TJSValueParam = record
     value: JSValue;
  end;

  TBluetoothCharacteristic = class;
  TBluetoothService = class;

  TReadIntEvent = procedure(Sender: TObject; AInt: integer) of object;
  TReadSingleEvent = procedure(Sender: TObject; ASingle: Single) of object;
  TReadDoubleEvent = procedure(Sender: TObject; ADouble: Double) of object;
  TReadStringEvent = procedure(Sender: TObject; AString: string) of object;
  TReadArrayEvent = procedure(Sender: TObject; AArray: array of Byte) of object;

  TGetDeviceErrorEvent = procedure(Sender: TObject; AError: TJSObject) of object;
  TGetDeviceObjectEvent = procedure(Sender:TObject; AObject: TJSObject) of object;
  TGetServiceObjectEvent = procedure(Sender: TObject; AOject: TJSObject) of object;
  TGetCharacteristicObjectEvent = procedure(Sender: TObject; AObject: TJSObject) of object;
  TGetDescriptorObjectEvent = procedure(Sender: TObject; AObject: TJSObject) of object;

  TGetServicesEvent = procedure(Sender: TObject; ServiceArray: array of TJSObject) of object;
  TGetCharacteristicsEvent = procedure(Sender: TObject; CharacteristicArray: array of TJSObject) of object;
  TGetDescriptorsEvent = procedure(Sender: TObject; DescriptorArray: array of TJSObject) of object;

  TGetNotifyIntEvent = procedure(Sender: TObject; AValue: integer) of object;
  TGetNotifyDoubleEvent = procedure(Sender: TObject; AValue: double) of object;
  TGetNotifyArrayEvent = procedure(Sender: TObject; AValue: array of byte) of object;

  TBTRefProc = reference to procedure;

  TBTServiceProc = reference to procedure(AService: TBluetoothService);
  TBTCharacteristicProc = reference to procedure(ACharacteristic: TBluetoothCharacteristic);
  TBTReadValueProc = reference to procedure(AValue: integer);
  TBTReadSingleProc = reference to procedure (AValue: single);
  TBTReadDoubleProc = reference to procedure(AValue: double);
  TBTReadArrayProc = reference to procedure (AValue: array of byte);

  { TBluetoothService }

  TBluetoothService = class(TPersistent)
  private
    FOnCharacteristic: TGetCharacteristicObjectEvent;
    FOnCharacteristics: TGetCharacteristicsEvent;
    FUUID: string;
    FOnService: TGetServiceObjectEvent;
  protected
    procedure DoGetCharacteristic(AObject: TJSObject);
    procedure DoGetCharacteristics(CharacteristicArray: array of TJSObject);
    procedure DoGetService(AObject: TJSObject);
    procedure SetServiceObject(AService: TJSObject);
    function CreateCharacteristic(AObject: TJSObject): TBluetoothCharacteristic;
  public
    constructor Create; overload;
    constructor Create(AService: TJSObject); overload;
    procedure GetCharacteristic(uuid: string); overload;
    procedure GetCharacteristic(uuid: string; proc: TBTCharacteristicProc); overload;
    procedure GetCharacteristic(uuid: integer); overload;
    procedure GetCharacteristic(uuid: integer; proc: TBTCharacteristicProc); overload;
    procedure GetCharacteristics;
  published
    property UUID: string read FUUID;
    property OnCharacteristic: TGetCharacteristicObjectEvent read FOnCharacteristic write FOnCharacteristic;
    property OnCharacteristics: TGetCharacteristicsEvent read FOnCharacteristics write FOnCharacteristics;
    property OnService: TGetServiceObjectEvent read FOnService write FOnService;
  end;

  { TBluetoothDevice }

  TBluetoothDevice = class(TPersistent)
  private
    FService: TBluetoothService;
    FDeviceObject: TJSObject;
    FOnService: TGetServiceObjectEvent;
    FOnDisconnect: TNotifyEvent;
    FOnConnect: TNotifyEvent;
    FDisconnectListenerAssigned: boolean;
    FOnServices: TGetServicesEvent;
  protected
    function DoHandleDisconnect(Event: TJSEvent): boolean;
    procedure DoGetService(AObject: TJSObject);
    procedure GetServiceAndConnect(uuid:string; proc: TBTServiceProc);
    procedure AssignDisconnectHandler;
    procedure DoConnect;
    procedure DoGetServices(ServiceArray: array of TJSObject);
    procedure SetDeviceObject(AObject: TJSObject);
    function CreateService(AObject: TJSObject): TBluetoothService;
  public
    constructor Create; overload;
    constructor Create(ADevice: TJSObject); overload;
    destructor Destroy; override;
    procedure GetService(uuid: string); overload;
    procedure GetService(uuid: string; proc: TBTServiceProc); overload;
    procedure GetServices;
    procedure Connect; overload;
    procedure Connect(proc: TBTRefProc); overload;
    function DisconnectDevice:boolean;
    function ReConnectDevice:boolean;
    function Connected: boolean;
    property Service: TBluetoothService read FService;
  published
    property OnConnect : TNotifyEvent read FOnConnect write FOnConnect;
    property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    property OnService: TGetServiceObjectEvent read FOnService write FOnService;
    property OnServices: TGetServicesEvent read FOnServices write FOnServices;
  end;

  { TBluetoothCharacteristic }

  TBluetoothCharacteristic = class(TPersistent)
  private
    FUUID: string;
    FOnReadValue: TReadIntEvent;
    FOnDescriptors: TGetDescriptorsEvent;
    FOnReadArray: TReadArrayEvent;
    FOnReadSingle: TReadSingleEvent;
    FOnReadString: TReadStringEvent;
    FOnReadDouble: TReadDoubleEvent;
    FOnDescriptor: TGetDescriptorObjectEvent;
    FOnStartNotify: TNotifyEvent;
    FOnHandleNotification: TNotifyEvent;
    FOnNotifyInt: TGetNotifyIntEvent;
    FOnNotifyDouble: TGetNotifyDoubleEvent;
    FOnNotifyArray: TGetNotifyArrayEvent;
    FOnWriteDone: TNotifyEvent;
  protected
    procedure DoGetDescriptor(AObject: TJSObject);
    procedure DoGetDescriptors(DescriptorArray: array of TJSObject);
    procedure DoRead(AValue:integer); virtual;
    procedure DoReadSingle(ASingle: single); virtual;
    procedure DoReadDouble(ADouble: double); virtual;
    procedure DoReadArray(AArray: array of byte); virtual;
    procedure DoReadString(a: array of byte); virtual;
    procedure DoStartNotify;
    procedure DoStopNotify;
    procedure DoNotify(AValue: integer); virtual;
    procedure DoNotifyDouble(AValue: double); virtual;
    procedure DoNotifyArray(AValue: array of byte); virtual;
    procedure AssignNotifyHandler;
    procedure UnAssignNotifyHandler;
    procedure DoHandleNotifications(Event: TJSEvent);
  public
    constructor Create(ACharacteristic: TJSObject);
    procedure GetDescriptor(uuid: string); overload;
    procedure GetDescriptor(uuid: integer); overload;
    procedure GetDescriptors;
    procedure StartNotify;
    procedure StopNotify;

    procedure Read; overload;
    procedure Read(proc: TBTReadValueProc); overload;
    procedure ReadSingle; overload;
    procedure ReadSingle(proc: TBTReadSingleProc); overload;
    procedure ReadDouble; overload;
    procedure ReadDouble(proc: TBTReadDoubleProc); overload;
    procedure ReadArray; overload;
    procedure ReadArray(proc: TBTReadArrayProc); overload;
    procedure ReadString;
    procedure WriteByte(b: byte);
    procedure WriteSmallInt(s: smallint);
    procedure WriteInt(i: integer);
    procedure WriteArray(a: array of byte);
    procedure WriteString(s: string);
    procedure WriteSingle(s: single);
    procedure WriteDouble(d: double);
  published
    property UUID: string read FUUID;
    property OnReadValue: TReadIntEvent read FOnReadValue write FOnReadValue;
    property OnReadSingle: TReadSingleEvent read FOnReadSingle write FOnReadSingle;
    property OnReadDouble: TReadDoubleEvent read FOnReadDouble write FOnReadDouble;
    property OnReadArray: TReadArrayEvent read FOnReadArray write FOnReadArray;
    property OnReadString: TReadStringEvent read FOnReadString write FOnReadString;
    property OnDescriptor: TGetDescriptorObjectEvent read FOnDescriptor write FOnDescriptor;
    property OnDescriptors: TGetDescriptorsEvent read FOnDescriptors write FOnDescriptors;
    property OnStartNotify: TNotifyEvent read FOnStartNotify write FOnStartNotify;
    property OnHandleNotification: TNotifyEvent read FOnHandleNotification write FOnHandleNotification;
    property OnNotifyInt: TGetNotifyIntEvent read FOnNotifyInt write FOnNotifyInt;
    property OnNotifyDouble: TGetNotifyDoubleEvent read FOnNotifyDouble write FOnNotifyDouble;
    property OnNotifyArray: TGetNotifyArrayEvent read FOnNotifyArray write FOnNotifyArray;
    property OnWriteDone: TNotifyEvent read FOnWriteDone write FOnWriteDone;
  end;

  { TBluetoothDescriptor }

  TBluetoothDescriptor = class(TPersistent)
  private
    FDescriptorObject: TJSObject;
    FUUID: string;
    FOnReadString: TReadStringEvent;
  protected
     procedure DoReadString(a: array of byte); virtual;
  public
    constructor Create(ADescriptor: TJSObject);
    procedure ReadString;
  published
    property UUID: string read FUUID;
    property OnReadString: TReadStringEvent read FOnReadString write FOnReadString;
  end;

  { TBluetooth }

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TWebBluetooth = class(TComponent)
  private
    FDevice: TBluetoothDevice;
    FDeviceName: string;
    FFilterService: TStringList;
    FOnDeviceObject: TGetDeviceObjectEvent;
    FOnDeviceError: TGetDeviceErrorEvent;
    procedure SetFilterService(const Value: TStringList);
  protected
    procedure DoGetDeviceObject(AObject: TJSObject); virtual;
    procedure DoGetDeviceError(AError: TJSObject); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HasBluetooth: boolean;
    function GetDevice: boolean; overload;
    function GetDevice(proc: TBTRefProc): boolean; overload;
    property Device: TBluetoothDevice read FDevice;
  published
    property DeviceName: string read FDeviceName write FDeviceName;
    property FilterService: TStringList read FFilterService write SetFilterService;
    property OnDeviceObject: TGetDeviceObjectEvent read FOnDeviceObject write FOnDeviceObject;
    property OnDeviceError: TGetDeviceErrorEvent read FOnDeviceError write FOnDeviceError;
  end;

implementation

{ TBluetooth }

constructor TWebBluetooth.Create(AOwner: TComponent);
begin
  inherited;
  FDevice := TBluetoothDevice.Create;
  FFilterService := TStringList.Create;
end;

destructor TWebBluetooth.Destroy;
begin
  FFilterService.Free;
  FDevice.Free;
  inherited;
end;

function TWebBluetooth.GetDevice: boolean;
begin
  Result := GetDevice(nil);
end;

function TWebBluetooth.GetDevice(proc: TBTRefProc): boolean;
begin
  Result := false;
  if not HasBluetooth then
    Exit;

  Result := true;
end;

function TWebBluetooth.HasBluetooth: boolean;
begin
  Result := true;
end;

procedure TWebBluetooth.SetFilterService(const Value: TStringList);
begin
  FFilterService.Assign(Value);
end;

procedure TWebBluetooth.DoGetDeviceError(AError: TJSObject);
begin
  if Assigned(OnDeviceError) then
    OnDeviceError(Self, AError);
end;

procedure TWebBluetooth.DoGetDeviceObject(AObject: TJSObject);
begin
  Device.SetDeviceObject(AObject);
  if Assigned (OnDeviceObject) then
    OnDeviceObject(Self,AObject);
end;

{------------------------------------------------------------------------------}

{ TBluetoothService }

constructor TBluetoothService.Create(AService: TJSObject);
begin
  inherited Create;
  SetServiceObject(AService);
end;

function TBluetoothService.CreateCharacteristic(
  AObject: TJSObject): TBluetoothCharacteristic;
begin
  Result := TBluetoothCharacteristic.Create(AObject);
end;

constructor TBluetoothService.Create;
begin
  Create(nil);
end;

procedure TBluetoothService.DoGetCharacteristic(AObject: TJSObject);
begin
  if Assigned(OnCharacteristic) then
    OnCharacteristic(Self, AObject);
end;

procedure TBluetoothService.DoGetCharacteristics(CharacteristicArray: array of TJSObject);
begin
  if Assigned(OnCharacteristics) then
    OnCharacteristics(Self, CharacteristicArray);
end;

procedure TBluetoothService.DoGetService(AObject: TJSObject);
begin
  if Assigned(OnService) then
    OnService(Self, AObject);
end;

procedure TBluetoothService.GetCharacteristic(uuid: string);
begin
  GetCharacteristic(uuid, nil);
end;

procedure TBluetoothService.GetCharacteristic(uuid: integer);
begin
  GetCharacteristic(uuid, nil);
end;

procedure TBluetoothService.GetCharacteristics;
begin
end;

procedure TBluetoothService.SetServiceObject(AService: TJSObject);
begin
end;

procedure TBluetoothService.GetCharacteristic(uuid: string; proc: TBTCharacteristicProc);
begin
end;

procedure TBluetoothService.GetCharacteristic(uuid: integer; proc: TBTCharacteristicProc);
begin
end;

{------------------------------------------------------------------------------}

{ TBluetoothCharacteristic }

procedure TBluetoothCharacteristic.AssignNotifyHandler;
begin
end;

constructor TBluetoothCharacteristic.Create(ACharacteristic: TJSObject);
begin
  inherited Create;
end;

procedure TBluetoothCharacteristic.DoGetDescriptor(AObject: TJSObject);
begin
  if Assigned(OnDescriptor) then
    OnDescriptor(Self,AObject);
end;

procedure TBluetoothCharacteristic.DoGetDescriptors(DescriptorArray: array of TJSObject);
begin
  if Assigned (OnDescriptors) then
    OnDescriptors(Self,DescriptorArray);
end;

procedure TBluetoothCharacteristic.DoHandleNotifications(Event: TJSEvent);
begin
  if Assigned(OnHandlenotification) then
    OnHandleNotification(Self);
end;

procedure TBluetoothCharacteristic.DoNotify(AValue: integer);
begin
  if Assigned(OnNotifyInt) then
    OnNotifyInt(Self, AValue);
end;

procedure TBluetoothCharacteristic.DoNotifyArray(AValue: array of byte);
begin
  if Assigned(OnNotifyArray) then
    OnNotifyArray(Self, AValue);
end;

procedure TBluetoothCharacteristic.DoNotifyDouble(AValue: double);
begin
  if Assigned(OnNotifyDouble) then
    OnNotifyDouble(Self, AValue);
end;

procedure TBluetoothCharacteristic.DostartNotify;
begin
  if Assigned(OnStartNotify) then
    OnStartNotify(Self);
end;

procedure TBluetoothCharacteristic.DoStopNotify;
begin
end;

procedure TBluetoothCharacteristic.DoRead(AValue: integer);
begin
  if Assigned(OnReadValue) then
    OnReadValue(Self, AValue);
end;

procedure TBluetoothCharacteristic.DoReadArray(AArray: array of byte);
begin
  if Assigned(OnReadArray) then
    OnReadArray(Self, AArray);
end;

procedure TBluetoothCharacteristic.DoReadDouble(ADouble: double);
begin
  if Assigned(OnReadDouble) then
    OnReadDouble(Self, ADouble);
end;

procedure TBluetoothCharacteristic.DoReadSingle(ASingle: single);
begin
  if Assigned(OnReadSingle) then
    OnReadSingle(Self, ASingle);
end;

procedure TBluetoothCharacteristic.DoReadString(a: array of byte);
begin
end;

procedure TBluetoothCharacteristic.GetDescriptor(uuid: integer);
begin
end;

procedure TBluetoothCharacteristic.GetDescriptors;
begin
end;

procedure TBluetoothCharacteristic.GetDescriptor(uuid: string);
begin
end;

procedure TBluetoothCharacteristic.Read(proc: TBTReadValueProc);
begin
end;

procedure TBluetoothCharacteristic.ReadArray;
begin
  ReadArray(nil);
end;

procedure TBluetoothCharacteristic.Read;
begin
  Read(nil);
end;

procedure TBluetoothCharacteristic.ReadArray(proc: TBTReadArrayProc);
begin
end;

procedure TBluetoothCharacteristic.ReadDouble;
begin
end;

procedure TBluetoothCharacteristic.ReadDouble(proc: TBTReadDoubleProc);
begin
end;

procedure TBluetoothCharacteristic.ReadSingle(proc: TBTReadSingleProc);
begin
end;

procedure TBluetoothCharacteristic.ReadSingle;
begin
end;

procedure TBluetoothCharacteristic.ReadString;
begin
end;

procedure TBluetoothCharacteristic.StartNotify;
begin
end;

procedure TBluetoothCharacteristic.StopNotify;
begin
end;

procedure TBluetoothCharacteristic.UnAssignNotifyHandler;
begin
end;

procedure TBluetoothCharacteristic.WriteArray(a: array of byte);
begin
end;

procedure TBluetoothCharacteristic.WriteByte(b: byte);
begin
end;

procedure TBluetoothCharacteristic.WriteDouble(d: double);
begin
end;

procedure TBluetoothCharacteristic.WriteSingle(s: single);
begin
end;

procedure TBluetoothCharacteristic.WriteSmallInt(s: smallint);
begin
end;

procedure TBluetoothCharacteristic.WriteString(s: string);
begin
end;

procedure TBluetoothCharacteristic.WriteInt(i: integer);
begin
end;

{------------------------------------------------------------------------------}

{ TBluetoothDevice }

procedure TBluetoothDevice.AssignDisconnectHandler;
begin
end;

procedure TBluetoothDevice.Connect;
begin
end;

procedure TBluetoothDevice.Connect(proc: TBTRefProc);
begin
end;


function TBluetoothDevice.Connected: boolean;
begin
  Result := false;
end;

constructor TBluetoothDevice.Create;
begin
  Create(nil);
end;

constructor TBluetoothDevice.Create(ADevice: TJSObject);
begin
  inherited Create;
  FDeviceObject := ADevice;
  FDisconnectListenerAssigned := false;
  FService := TBluetoothService.Create;
end;

function TBluetoothDevice.CreateService(AObject: TJSObject): TBluetoothService;
begin
  Result := TBluetoothService.Create(AObject);
end;

destructor TBluetoothDevice.Destroy;
begin
  FService.Free;
  inherited;
end;

function TBluetoothDevice.DisconnectDevice: boolean;
begin
  Result := false;
end;

procedure TBluetoothDevice.SetDeviceObject(AObject: TJSObject);
begin
  FDeviceObject := AObject;
end;

function TBluetoothDevice.DoHandleDisconnect(Event: TJSEvent): boolean;
begin
  Result := true;
end;

procedure TBluetoothDevice.DoConnect;
begin
end;

procedure TBluetoothDevice.DoGetService(AObject: TJSObject);
begin
end;


procedure TBluetoothDevice.DoGetServices(ServiceArray: array of TJSObject);
begin
 if Assigned(OnServices) then
    OnServices(Self, ServiceArray);
end;

procedure TBluetoothDevice.GetService(uuid: string);
begin
  GetService(uuid,nil);
end;

procedure TBluetoothDevice.GetService(uuid: string; proc: TBTServiceProc);
begin
end;

procedure TBluetoothDevice.GetServiceAndConnect(uuid:string; proc: TBTServiceProc);
begin
end;

procedure TBluetoothDevice.GetServices;
begin
end;

function TBluetoothDevice.ReConnectDevice: boolean;
begin
  Result := true;
end;

{------------------------------------------------------------------------------}

{ TBluetoothDescriptor }

constructor TBluetoothDescriptor.Create(ADescriptor: TJSObject);
begin
  inherited Create;
  FDescriptorObject := ADescriptor;
end;

procedure TBluetoothDescriptor.DoReadString(a: array of byte);
begin
end;

procedure TBluetoothDescriptor.ReadString;
begin
end;

end.
